#ifdef WIN32
#include <windows.h>
#define XRaiseWindow newXRaiseWindow
#endif

#include <assert.h>
/* #include "busy.h" */

#include "tkPlatDecls.h"

#define BUSY_THREAD_KEY "BLT Busy Data"

/* Forward declarations */
void DestroyBusy _ANSI_ARGS_((DestroyData dataPtr));

void
ShowBusyWindow(busyPtr) 
    Busy *busyPtr;
{
    if (busyPtr->busyTkwin != NULL) {
	Tk_MapWindow(busyPtr->busyTkwin);
	/* 
	 * Always restack the busy window, just in case new sibling
	 * windows have been created. Can't use Tk_RestackWindow 
	 * because it doesn't work under Win32.
	 */
	XRaiseWindow(Tk_Display(busyPtr->busyTkwin), 
	     Tk_WindowId(busyPtr->busyTkwin));
    }
#ifdef WIN32
    {
	POINT point;
	/*
	 * Under Win32, cursors aren't associated with windows.  Tk
	 * fakes this by watching Motion events on its windows.  So Tk
	 * will automatically change the cursor when the pointer
	 * enters the Busy window.  But Windows doesn't immediately
	 * change the cursor; it waits for the cursor position to
	 * change or a system call.  We need to change the cursor
 	 * before the application starts processing, so set the cursor
	 * position redundantly back to the current position.
	 */
	GetCursorPos(&point);
	SetCursorPos(point.x, point.y);
    }
#endif /* WIN32 */
}

void
HideBusyWindow(busyPtr)
    Busy *busyPtr;
{
    if (busyPtr->busyTkwin != NULL) {
	Tk_UnmapWindow(busyPtr->busyTkwin); 
    }
#ifdef WIN32
    {
	POINT point;
	/*
	 * Under Win32, cursors aren't associated with windows.  Tk
	 * fakes this by watching Motion events on its windows.  So Tk
	 * will automatically change the cursor when the pointer
	 * enters the Busy window.  But Windows doesn't immediately
	 * change the cursor; it waits for the cursor position to
	 * change or a system call.  We need to change the cursor
	 * before the application starts processing, so set the cursor
	 * position redundantly back to the current position.
	 */
	GetCursorPos(&point);
	SetCursorPos(point.x, point.y);
    }
#endif /* WIN32 */
}

/*
 *----------------------------------------------------------------------
 *
 * BusyEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for events on
 *	the busy window itself.  We're only concerned with destroy
 *	events.
 *
 *	It might be necessary (someday) to watch resize events.  Right
 *	now, I don't think there's any point in it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When a busy window is destroyed, all internal structures
 * 	associated with it released at the next idle point.
 *
 *----------------------------------------------------------------------
 */
static void
BusyEventProc(clientData, eventPtr)
    ClientData clientData;	/* Busy window record */
    XEvent *eventPtr;		/* Event which triggered call to routine */
{
    Busy *busyPtr = (Busy *)clientData;

    if (eventPtr->type == DestroyNotify) {
	busyPtr->busyTkwin = NULL;
	Tcl_EventuallyFree((ClientData)busyPtr, DestroyBusy);
    }
}

/*
 * ----------------------------------------------------------------------------
 *
 * BusyCustodyProc --
 *
 *	This procedure is invoked when the busy window has been stolen
 *	by another geometry manager.  The information and memory
 *	associated with the busy window is released. I don't know why
 *	anyone would try to pack a busy window, but this should keep
 *	everything sane, if it is.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The Busy structure is freed at the next idle point.
 *
 * ----------------------------------------------------------------------------
 */
/* ARGSUSED */
void
BusyCustodyProc(clientData, tkwin)
    ClientData clientData;	/* Information about the busy window. */
    Tk_Window tkwin;		/* Not used. */
{
    Busy *busyPtr = (Busy *)clientData;

    Tk_DeleteEventHandler(busyPtr->busyTkwin, StructureNotifyMask, 
	BusyEventProc, (ClientData)busyPtr);
    HideBusyWindow(busyPtr);
    busyPtr->busyTkwin = NULL;
    Tcl_EventuallyFree((ClientData)busyPtr, DestroyBusy);
}

/*
 * ----------------------------------------------------------------------------
 *
 * BusyGeometryProc --
 *
 *	This procedure is invoked by Tk_GeometryRequest for busy
 *	windows.  Busy windows never request geometry, so it's
 *	unlikely that this routine will ever be called.  The routine
 *	exists simply as a place holder for the GeomProc in the
 *	Geometry Manager structure.
 *
 * Results:
 *	None.
 *
 * ----------------------------------------------------------------------------
 */
/* ARGSUSED */
void
BusyGeometryProc(clientData, tkwin)
    ClientData clientData;	/* Information about window that got new
				 * preferred geometry.  */
    Tk_Window tkwin;		/* Other Tk-related information about the
			         * window. */
{
    /* Should never get here */
}

/*
 * ------------------------------------------------------------------
 *
 * RefWinEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for the
 *	following events on the reference window.  If the reference and
 *	parent windows are the same, only the first event is
 *	important.
 *
 *	   1) ConfigureNotify  - The reference window has been resized or
 *				 moved.  Move and resize the busy window
 *				 to be the same size and position of the
 *				 reference window.
 *
 *	   2) DestroyNotify    - The reference window was destroyed. Destroy
 *				 the busy window and the free resources
 *				 used.
 *
 *	   3) MapNotify	       - The reference window was (re)shown. Map the
 *				 busy window again.
 *
 *	   4) UnmapNotify      - The reference window was hidden. Unmap the
 *				 busy window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the reference window gets deleted, internal structures get
 *	cleaned up.  When it gets resized, the busy window is resized
 *	accordingly. If it's displayed, the busy window is displayed. And
 *	when it's hidden, the busy window is unmapped.
 *
 * -------------------------------------------------------------------
 */
static void
RefWinEventProc(clientData, eventPtr)
    ClientData clientData;	/* Busy window record */
    register XEvent *eventPtr;	/* Event which triggered call to routine */
{
    register Busy *busyPtr = (Busy *)clientData;

    switch (eventPtr->type) {
    case ReparentNotify:
    case DestroyNotify:

	/*
	 * Arrange for the busy structure to be removed at a proper time.
	 */

	Tcl_EventuallyFree((ClientData)busyPtr, DestroyBusy);
	break;

    case ConfigureNotify:
	if ((busyPtr->width != Tk_Width(busyPtr->client)) ||
	    (busyPtr->height != Tk_Height(busyPtr->client)) ||
	    (busyPtr->x != Tk_X(busyPtr->client)) ||
	    (busyPtr->y != Tk_Y(busyPtr->client))) {
	    int x, y;

	    busyPtr->width = Tk_Width(busyPtr->client);
	    busyPtr->height = Tk_Height(busyPtr->client);
	    busyPtr->x = Tk_X(busyPtr->client);
	    busyPtr->y = Tk_Y(busyPtr->client);

	    x = y = 0;
	    if (busyPtr->parent != busyPtr->client) {
		Tk_Window tkwin;

		for (tkwin = busyPtr->client; (tkwin != NULL) &&
			 (!Tk_IsTopLevel(tkwin)); tkwin = Tk_Parent(tkwin)) {
		    x += Tk_X(tkwin) + Tk_Changes(tkwin)->border_width;
		    y += Tk_Y(tkwin) + Tk_Changes(tkwin)->border_width;
		    if (tkwin == busyPtr->parent) {
			break;
		    }
		}
	    }
	    if (busyPtr->busyTkwin != NULL) {
		Tk_MoveResizeWindow(busyPtr->busyTkwin, x, y, busyPtr->width,
		    busyPtr->height);
		if (busyPtr->isBusy) {
		    ShowBusyWindow(busyPtr);
		}
	    }
	}
	break;

    case MapNotify:
	if ((busyPtr->parent != busyPtr->client) && (busyPtr->isBusy)) {
	    ShowBusyWindow(busyPtr);
	}
	break;

    case UnmapNotify:
	if (busyPtr->parent != busyPtr->client) {
	    HideBusyWindow(busyPtr);
	}
	break;
    }
}

/*
 * ------------------------------------------------------------------
 *
 * ConfigureBusy --
 *
 *	This procedure is called from the Tk event dispatcher. It
 * 	releases X resources and memory used by the busy window and
 *	updates the internal hash table.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory and resources are released and the Tk event handler
 *	is removed.
 *
 * -------------------------------------------------------------------
 */
int
ConfigureBusy(interp, busyPtr, argc, argv)
    Tcl_Interp *interp;
    Busy *busyPtr;
    int argc;
    char **argv;
{
    Tk_Cursor oldCursor;

    oldCursor = busyPtr->cursor;
    if (Tk_ConfigureWidget(interp, busyPtr->client, configSpecs, argc, argv,
	    (char *)busyPtr, 0) != TCL_OK) {
	return TCL_ERROR;
    }
    if (busyPtr->cursor != oldCursor) {
	if (busyPtr->cursor == None) {
	    Tk_UndefineCursor(busyPtr->busyTkwin);
	} else {
	    Tk_DefineCursor(busyPtr->busyTkwin, busyPtr->cursor);
	}
    }
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * CreateBusy --
 *
 *	Creates a child transparent window that obscures its parent
 *	window thereby effectively blocking device events.  The size
 *	and position of the busy window is exactly that of the reference
 *	window.
 *
 *	We want to create sibling to the window to be blocked.  If the
 *	busy window is a child of the window to be blocked, Enter/Leave
 *	events can sneak through.  Futhermore under WIN32, messages of
 *	transparent windows are sent directly to the parent.  The only
 *	exception to this are toplevels, since we can't make a sibling.
 *	Fortunately, toplevel windows rarely receive events that need
 *	blocking.
 *
 * Results:
 *	Returns a pointer to the new busy window structure.
 *
 * Side effects:
 *	When the busy window is eventually displayed, it will screen
 *	device events (in the area of the reference window) from reaching
 *	its parent window and its children.  User feed back can be
 *	achieved by changing the cursor.
 *
 * -------------------------------------------------------------------
 */
Busy *
CreateBusy(interp, client)
    Tcl_Interp *interp;		/* Interpreter to report error to */
    Tk_Window client;		/* Window hosting the busy window */
{
    Busy *busyPtr;
    int length;
    char *fmt, *name;
    Tk_Window busyTkwin;
    Window parentWindow;
    Tk_Window child, parent, FirstChild(), NextChild();
    Tk_FakeWin *winPtr;
    int x, y;

    busyPtr = (Busy *)calloc(1, sizeof(Busy));
    assert(busyPtr);
    x = y = 0;
    length = strlen(Tk_Name(client));
    name = (char *)malloc(length + 6);
    if (Tk_IsTopLevel(client)) {
	fmt = "_Busy";		/* Child */
	parent = client;
    } else {
	Tk_Window tkwin;

	fmt = "%s_Busy";	/* Sibling */
	parent = Tk_Parent(client);
	for (tkwin = client; (tkwin != NULL) && (!Tk_IsTopLevel(tkwin)); 
	     tkwin = Tk_Parent(tkwin)) {
	    x += Tk_X(tkwin) + Tk_Changes(tkwin)->border_width;
	    y += Tk_Y(tkwin) + Tk_Changes(tkwin)->border_width;
	    if (tkwin == parent) {
		break;
	    }
	}
    }
    for (child = FirstChild(parent); child != NULL; child = NextChild(child)) {
        Tk_MakeWindowExist(child);
    }
    sprintf(name, fmt, Tk_Name(client));
    busyTkwin = Tk_CreateWindow(interp, parent, name, (char *)NULL);
    free((char *)name);

    if (busyTkwin == NULL) {
	return NULL;
    }
    Tk_MakeWindowExist(client);
    busyPtr->display = Tk_Display(client);
    busyPtr->client = client;
    busyPtr->parent = parent;
    busyPtr->interp = interp;
    busyPtr->width = Tk_Width(client);
    busyPtr->height = Tk_Height(client);
    busyPtr->x = Tk_X(client);
    busyPtr->y = Tk_Y(client);
    busyPtr->cursor = None;
    busyPtr->busyTkwin = busyTkwin;
    busyPtr->isBusy = FALSE;
    Tk_SetClass(busyTkwin, "Busy");
    SetWindowInstanceData(busyTkwin, (ClientData)busyPtr);
    winPtr = (Tk_FakeWin *) client;
    if (winPtr->flags & TK_REPARENTED) {
	/*
	 * This works around a bug in the implementation of menubars
	 * for non-MacIntosh window systems (Win32 and X11).  Tk
	 * doesn't reset the pointers to the parent window when the
	 * menu is reparented (winPtr->parentPtr points to the
	 * wrong window). We get around this by determining the parent
	 * via the native API calls.
	 */
#ifdef WIN32
	{
	    HWND hWnd;
	    RECT region;

	    hWnd = GetParent(Tk_GetHWND(Tk_WindowId(client)));
	    parentWindow = (Window) hWnd;
	    if (GetWindowRect(hWnd, &region)) {
            busyPtr->width = region.right - region.left;
            busyPtr->height = region.bottom - region.top;
	    }
	}
#else
	parentWindow = GetParent(Tk_Display(client), Tk_WindowId(client));
#endif
    } else {
        parentWindow = Tk_WindowId(parent);
#ifdef WIN32
        parentWindow = (Window) Tk_GetHWND(parentWindow);
#endif
    }
    MakeTransparentWindowExist(busyTkwin, parentWindow, TRUE);

    Tk_MoveResizeWindow(busyTkwin, x, y, busyPtr->width, busyPtr->height);

    /* Only worry if the busy window is destroyed.  */
    Tk_CreateEventHandler(busyTkwin, StructureNotifyMask, BusyEventProc,
                                (ClientData)busyPtr);
    /*
     * Indicate that the busy window's geometry is being managed.
     * This will also notify us if the busy window is ever packed.
     */
    Tk_ManageGeometry(busyTkwin, &busyMgrInfo, (ClientData)busyPtr);
    if (busyPtr->cursor != None) {
        Tk_DefineCursor(busyTkwin, busyPtr->cursor);
    }
    /* Track the reference window to see if it is resized or destroyed.  */
    Tk_CreateEventHandler(client, StructureNotifyMask, RefWinEventProc,
                                    (ClientData)busyPtr);
    return busyPtr;
}

/*
 * ------------------------------------------------------------------
 *
 * DestroyBusy --
 *
 *	This procedure is called from the Tk event dispatcher. It
 *	releases X resources and memory used by the busy window and
 *	updates the internal hash table.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory and resources are released and the Tk event handler
 *	is removed.
 *
 * -------------------------------------------------------------------
 */
void
DestroyBusy(dataPtr)
    DestroyData dataPtr;	/* Busy window structure record */
{
    Busy *busyPtr = (Busy *)dataPtr;

    Tk_FreeOptions(configSpecs, (char *)busyPtr, busyPtr->display, 0);
    if (busyPtr->hashPtr != NULL) {
	Tcl_DeleteHashEntry(busyPtr->hashPtr);
    }
    Tk_DeleteEventHandler(busyPtr->client, StructureNotifyMask,
	RefWinEventProc, (ClientData)busyPtr);
    if (busyPtr->busyTkwin != NULL) {
	Tk_DeleteEventHandler(busyPtr->busyTkwin, StructureNotifyMask,
	    BusyEventProc, (ClientData)busyPtr);
	Tk_ManageGeometry(busyPtr->busyTkwin, (Tk_GeomMgr *) NULL,
	    (ClientData)busyPtr);
	Tk_DestroyWindow(busyPtr->busyTkwin);
    }
    free((char *)busyPtr);
}

/*
 * ------------------------------------------------------------------
 *
 * GetBusy --
 *
 *	Returns the busy window structure associated with the reference
 *	window, keyed by its path name.  The clientData argument is
 *	the main window of the interpreter, used to search for the
 *	reference window in its own window hierarchy.
 *
 * Results:
 *	If path name represents a reference window with a busy window, a
 *	pointer to the busy window structure is returned. Otherwise,
 *	NULL is returned and an error message is left in
 *	interp->result.
 *
 * -------------------------------------------------------------------
 */
int
GetBusy(dataPtr, interp, pathName, busyPtrPtr)
    ThreadData *dataPtr;		/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    char *pathName;		/* Path name of parent window */
    Busy **busyPtrPtr;		/* Will contain address of busy window if
				 * found. */
{
    Tcl_HashEntry *hPtr;
    Tk_Window tkwin;

    tkwin = Tk_NameToWindow(interp, pathName, Tk_MainWindow(interp));
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    hPtr = Tcl_FindHashEntry(&(dataPtr->busyTable), (char *)tkwin);
    if (hPtr == NULL) {
	Tcl_AppendResult(interp, "can't find busy window \"", pathName, "\"",
	    (char *)NULL);
	return TCL_ERROR;
    }
    *busyPtrPtr = ((Busy *)Tcl_GetHashValue(hPtr));
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * HoldBusy --
 *
 *	Creates (if necessary) and maps a busy window, thereby
 *	preventing device events from being be received by the parent
 *	window and its children.
 *
 * Results:
 *	Returns a standard TCL result. If path name represents a busy
 *	window, it is unmapped and TCL_OK is returned. Otherwise,
 *	TCL_ERROR is returned and an error message is left in
 *	interp->result.
 *
 * Side effects:
 *	The busy window is created and displayed, blocking events from
 *	the parent window and its children.
 *
 * -------------------------------------------------------------------
 */
static int
HoldBusy(dataPtr, interp, argc, argv)
    ThreadData *dataPtr;		/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;		/* Window name and option pairs */
{
    Tk_Window tkwin;
    Tcl_HashEntry *hPtr;
    Busy *busyPtr;
    int isNew;
    int result;


    tkwin = Tk_NameToWindow(interp, argv[0], Tk_MainWindow(interp));
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    hPtr = Tcl_CreateHashEntry(&(dataPtr->busyTable), (char *)tkwin, &isNew);
    if (isNew) {
	busyPtr = (Busy *)CreateBusy(interp, tkwin);
	if (busyPtr == NULL) {
	    return TCL_ERROR;
	}
	Tcl_SetHashValue(hPtr, (char *)busyPtr);
	busyPtr->hashPtr = hPtr;
    } else {
	busyPtr = (Busy *)Tcl_GetHashValue(hPtr);
    }
    result = ConfigureBusy(interp, busyPtr, argc - 1, argv + 1);
    /* Don't map the busy window unless the reference window is also
     * displayed */
    if (Tk_IsMapped(busyPtr->client)) {
	ShowBusyWindow(busyPtr);
    } else {
	HideBusyWindow(busyPtr);
    }
    busyPtr->isBusy = TRUE;
    return result;
}

/*
 * ------------------------------------------------------------------
 *
 * StatusOp --
 *
 *	Returns the status of the busy window; whether it's blocking
 *	events or not.
 *
 * Results:
 *	Returns a standard TCL result. If path name represents a busy
 *	window, the status is returned via interp->result and TCL_OK
 *	is returned. Otherwise, TCL_ERROR is returned and an error
 *	message is left in interp->result.
 *
 * -------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
StatusOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report error to */
    int argc;			/* Not used. */
    char **argv;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Busy *busyPtr;

    if (GetBusy(dataPtr, interp, argv[2], &busyPtr) != TCL_OK) {
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData)busyPtr);
    Tcl_SetResult(interp, busyPtr->isBusy ? "1" : "0", TCL_STATIC);
    Tcl_Release((ClientData)busyPtr);
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * ForgetOp --
 *
 *	Destroys the busy window associated with the reference window and
 *	arranges for internal resources to the released when they're
 *	not being used anymore.
 *
 * Results:
 *	Returns a standard TCL result. If path name represents a busy
 *	window, it is destroyed and TCL_OK is returned. Otherwise,
 *	TCL_ERROR is returned and an error message is left in
 *	interp->result.
 *
 * Side effects:
 *	The busy window is removed.  Other related memory and resources
 *	are eventually released by the Tk dispatcher.
 *
 * -------------------------------------------------------------------
 */
static int
ForgetOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Busy *busyPtr;
    register int i;

    for (i = 2; i < argc; i++) {
	if (GetBusy(dataPtr, interp, argv[i], &busyPtr) != TCL_OK) {
	    return TCL_ERROR;
	}
	/* Unmap the window even though it will be soon destroyed */
	HideBusyWindow(busyPtr);
	Tcl_EventuallyFree((ClientData)busyPtr, DestroyBusy);
    }
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * ReleaseOp --
 *
 *	Unmaps the busy window, thereby permitting device events
 *	to be received by the parent window and its children.
 *
 * Results:
 *	Returns a standard TCL result. If path name represents a busy
 *	window, it is unmapped and TCL_OK is returned. Otherwise,
 *	TCL_ERROR is returned and an error message is left in
 *	interp->result.
 *
 * Side effects:
 *	The busy window is hidden, allowing the parent window and
 *	its children to receive events again.
 *
 * -------------------------------------------------------------------
 */
static int
ReleaseOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Busy *busyPtr;
    int i;

    for (i = 2; i < argc; i++) {
	if (GetBusy(dataPtr, interp, argv[i], &busyPtr) != TCL_OK) {
	    return TCL_ERROR;
	}
	HideBusyWindow(busyPtr);
	busyPtr->isBusy = FALSE;
    }
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * WindowsOp --
 *
 *	Reports the names of all widgets with busy windows attached to
 *	them, matching a given pattern.  If no pattern is given, all
 *	busy widgets are listed.
 *
 * Results:
 *	Returns a TCL list of the names of the widget with busy windows
 *	attached to them, regardless if the widget is currently busy
 *	or not.
 *
 * -------------------------------------------------------------------
 */
static int
WindowsOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch cursor;
    Busy *busyPtr;

    for (hPtr = Tcl_FirstHashEntry(&(dataPtr->busyTable), &cursor);
	hPtr != NULL; hPtr = Tcl_NextHashEntry(&cursor)) {
	busyPtr = (Busy *)Tcl_GetHashValue(hPtr);
	if (busyPtr->interp == interp) {
	    if ((argc != 3) ||
		(Tcl_StringMatch(Tk_PathName(busyPtr->client), argv[2]))) {
		Tcl_AppendElement(interp, Tk_PathName(busyPtr->client));
	    }
	}
    }
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 *
 * IsBusyOp --
 *
 *	Reports the names of all widgets with busy windows attached to
 *	them, matching a given pattern.  If no pattern is given, all
 *	busy widgets are listed.
 *
 * Results:
 *	Returns a TCL list of the names of the widget with busy windows
 *	attached to them, regardless if the widget is currently busy
 *	or not.
 *
 * -------------------------------------------------------------------
 */
static int
IsBusyOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch cursor;
    Busy *busyPtr;

    for (hPtr = Tcl_FirstHashEntry(&(dataPtr->busyTable), &cursor);
	hPtr != NULL; hPtr = Tcl_NextHashEntry(&cursor)) {
	busyPtr = (Busy *)Tcl_GetHashValue(hPtr);
	if (busyPtr->interp == interp) {
	    if ((busyPtr->isBusy) && ((argc != 3) ||
		    (Tcl_StringMatch(Tk_PathName(busyPtr->client), argv[2])))) {
		Tcl_AppendElement(interp, Tk_PathName(busyPtr->client));
	    }
	}
    }
    return TCL_OK;
}


/* ARGSUSED*/
static int
CgetOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;		/* Widget pathname and option switch */
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Busy *busyPtr;
    int result;

    if (GetBusy(dataPtr, interp, argv[2], &busyPtr) != TCL_OK) {
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData)busyPtr);
    result = Tk_ConfigureValue(interp, busyPtr->client, configSpecs,
	(char *)busyPtr, argv[3], 0);
    Tcl_Release((ClientData)busyPtr);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureOp --
 *
 *	This procedure is called to process an argv/argc list in order
 *	to configure (or reconfigure) a busy window.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information get set for busyPtr; old resources
 *	get freed, if there were any.  The busy window destroyed and
 *	recreated in a new parent window.
 *
 *----------------------------------------------------------------------
 */
static int
ConfigureOp(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter to report errors to */
    int argc;
    char **argv;		/* Reference window path name and options */
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Busy *busyPtr;
    int result;

    if (GetBusy(dataPtr, interp, argv[2], &busyPtr) != TCL_OK) {
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData)busyPtr);
    if (argc == 3) {
	result = Tk_ConfigureInfo(interp, busyPtr->client, configSpecs,
	    (char *)busyPtr, (char *)NULL, 0);
    } else if (argc == 4) {
	result = Tk_ConfigureInfo(interp, busyPtr->client, configSpecs,
	    (char *)busyPtr, argv[3], 0);
    } else {
	result = ConfigureBusy(interp, busyPtr, argc - 3, argv + 3);
    }
    Tcl_Release((ClientData)busyPtr);
    return result;
}

/*
 * -----------------------------------------------------------------------
 *
 * BusyInterpDeleteProc --
 *
 *	This is called when the interpreter hosting the "busy" command 
 *	is destroyed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Destroys all the hash table managing the busy windows.
 *
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static void
BusyInterpDeleteProc(clientData, interp)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;
{
    ThreadData *dataPtr = (ThreadData *)clientData;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch cursor;
    Busy *busyPtr;

    for (hPtr = Tcl_FirstHashEntry(&(dataPtr->busyTable), &cursor);
	 hPtr != NULL; hPtr = Tcl_NextHashEntry(&cursor)) {
	busyPtr = (Busy *)Tcl_GetHashValue(hPtr);
	busyPtr->hashPtr = NULL;
	DestroyBusy((DestroyData)busyPtr);
    }
    Tcl_DeleteHashTable(&(dataPtr->busyTable));
    Tcl_DeleteAssocData(interp, BUSY_THREAD_KEY);
    free((char *)dataPtr);
}

#ifdef NOTDEF

/*
 *--------------------------------------------------------------
 *
 * Busy Sub-command specification:
 *
 *	- Name of the sub-command.
 *	- Minimum number of characters needed to unambiguously
 *        recognize the sub-command.
 *	- Pointer to the function to be called for the sub-command.
 *	- Minimum number of arguments accepted.
 *	- Maximum number of arguments accepted.
 *	- String to be displayed for usage (arguments only).
 *
 *--------------------------------------------------------------
 */
static Blt_OpSpec busyOps[] =
{
    {"cget", 2, (Blt_OpProc)CgetOp, 4, 4, "window option",},
    {"configure", 2, (Blt_OpProc)ConfigureOp, 3, 0,
	"window ?options?...",},
    {"forget", 1, (Blt_OpProc)ForgetOp, 2, 0, "?window?...",},
    {"hold", 3, (Blt_OpProc)HoldOp, 3, 0,
	"window ?options?... ?window options?...",},
    {"isbusy", 1, (Blt_OpProc)IsBusyOp, 2, 3, "?pattern?",},
    {"release", 1, (Blt_OpProc)ReleaseOp, 2, 0, "?window?...",},
    {"status", 1, (Blt_OpProc)StatusOp, 3, 3, "window",},
    {"windows", 1, (Blt_OpProc)WindowsOp, 2, 3, "?pattern?",},
};
static int nBusyOps = sizeof(busyOps) / sizeof(Blt_OpSpec);
#endif

#ifdef notdef
/*
 *----------------------------------------------------------------------
 *
 * BusyCmd --
 *
 *	This procedure is invoked to process the "busy" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */
/* static int */
int
BusyCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Thread-specific data. */
    Tcl_Interp *interp;		/* Interpreter associated with command */
    int argc;
    char **argv;
{
    Blt_OpProc opProc;
    int result;
    
    if ((argc > 1) && (argv[1][0] == '.')) {
        return HoldOp(clientData, interp, argc, argv);
    }
    opProc = (Blt_OpProc) Blt_GetOperation(interp, nBusyOps, busyOps,
                BLT_OPER_ARG1, argc, argv);
    if (opProc == NULL) {
	return TCL_ERROR;
    }
    result = (*opProc) (clientData, interp, argc, argv);
    return result;
}
#endif

ThreadData *
GetBusyCmdData(interp)
     Tcl_Interp *interp;
{
    ThreadData *dataPtr;
    Tcl_InterpDeleteProc *proc;

    dataPtr = (ThreadData *)Tcl_GetAssocData(interp, BUSY_THREAD_KEY, &proc);
    if (dataPtr == NULL) {
	dataPtr = (ThreadData *)malloc(sizeof(ThreadData));
	assert(dataPtr);
	Tcl_SetAssocData(interp, BUSY_THREAD_KEY, BusyInterpDeleteProc, 
		(ClientData)dataPtr);
	Tcl_InitHashTable(&(dataPtr->busyTable), TCL_ONE_WORD_KEYS);
    }
    return dataPtr;
}
